from .Token import Token

class Span():

    __slots__ = ["ref_document", "sentence_idx", "tokens_idx", "cluster_idx"]

    def __init__(self, span_data, ref_document):
        """Class for coreference spans

        Args:
            span_data (dict): Span data containing sentence, tokens and cluster indexes. 
            ref_document (Document): Reference to the Document object.
        """        
        self.sentence_idx = span_data['sentence_index']
        self.tokens_idx = span_data['token_indexes']
        self.cluster_idx = span_data['cluster_index']
        self.ref_document = ref_document

    @property    
    def tokens(self) -> Token:
        return self.ref_document.sentences[self.sentence_idx].tokens[self.tokens_idx[0]: self.tokens_idx[-1] + 1]

    def __str__(self):
        return ' '.join([s.str for s in self.tokens])

    def __repr__(self):
        return str(self)

    def get_attributes(self, property):
        return [getattr(s, property) for s in self.tokens]


from .utils import flatten_lst, StrProperty, ListProperty, DictProperty, IntProperty
from .utils import GLOBAL, DOC, SENT, SUB, TOK
from .TextChunk import TextChunk
from .Subsentence import Subsentence
from .Token import Token
from .Span import Span
from .Cluster import Cluster

class Sentence(TextChunk):
    __slots__ = ("data", "n", "max", "_id", "ref_document")

    def __init__(self, data_sentence, idx=0, ref_document=None):
        """ Class that represents a sentence.

        Args:
            data_sentence (dict): dict with sentence data.
            idx (int, optional): Index of the sentence in the document. Defaults to 0.
            ref_document (Document, optional): Reference to the document. Defaults to None.
        """
        super(Sentence, self).__init__()
        self.data = data_sentence
        self.max = len(self.data.get('detail', []))
        self._ner_to_detail()
        self._id = idx
        self.ref_document = ref_document

    @property
    def id(self) -> int:
        return self._id

    @property
    def idx(self) -> int:
        return self.id

    def _ner_to_detail(self):
        """ Inject NER data into the detail key to be used by Token."""
        for m in self.data.get('ml_ner', []):
            if len(self.data.get('detail', [])) > m['index']:
                self.data['detail'][m['index']]['value'] = m['value']
                self.data['detail'][m['index']]['type'] = m['type']

    def __repr__(self):
        return self.str

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> Token:
        if self.n < self.max:
            self.n += 1
            return Token(
                self.data.get('detail', [])[self.n - 1], 
                self.n -1, 
                self.data.get('source_pure', self.data.get('source', None)), 
                self)
        else:
            raise StopIteration

    @ListProperty
    def sentences(self) -> list:
        return [self]

    def _get_subsentence(self, _id, _idx) -> Subsentence:
        """ Divides sentence information to each subsentence."""
        data = {}
        data['detail'] = self.data.get('detail', [])[_idx['start_id']:_idx['end_id'] + 1]
        data['source'] = ' '.join([k.get('source', '') if k.get('source', '') else '' for k in data['detail']])
        data['source_pure'] = self.data.get('source_pure', '')
        if len(self.data.get('sentiment', {}).get('subsentences', [])) > _id:
            data['sentiment'] = self.data['sentiment'].get('subsentences', [])[_id]
        if len(self.data.get('emotion', {}).get('subsentences', [])) > _id:
            data['emotion'] = self.data['emotion'].get('subsentences', [])[_id]
        if len(self.data.get('ml_sentiment', {}).get('subsentence', [])) > _id:
            data['ml_sentiment'] = {'sentence': self.data['ml_sentiment']['sentence'], 'subsentence': {'value': self.data['ml_sentiment']['subsentence'][_id]}}
        if len(self.data.get('ml_emotion', {}).get('subsentence', [])) > _id:
            data['ml_emotion'] = {'sentence': self.data['ml_emotion']['sentence'], 'subsentence': self.data['ml_emotion']['subsentence'][_id]}
        return Subsentence(data)

    @ListProperty
    def tokens(self) -> list:
        return [Token(s, i, self.data.get('source_pure', self.data.get('source', None)), self) for i, s in enumerate(self.data.get('detail', []))]

    @ListProperty
    def subsentences(self) -> list:
        return [self._get_subsentence(_id, idx) for _id, idx in enumerate(self.data.get('subsentences', []))]

    @StrProperty
    def str(self) -> str:
        return self.data.get('source', None)

    @StrProperty
    def original_text(self) -> str:
        return self.data.get('source_pure', self.data.get('source', None))

    @ListProperty
    def synthesis(self) -> list: ## for compatibility purpose
        return self.data.get('detail', [])

    @ListProperty
    def detail(self) -> list:
        return self.data.get('detail', [])

    @ListProperty
    def emotion(self) -> list:
        return [(k, round(v,4)) for k,v in self.data.get('emotion', {}).get('values', {}).items() if v != 0]

    @DictProperty
    def emotion_ml(self) -> list:
        if self.__class__.__name__ == 'Sentence':
            return [(dic['type'], round(dic['value'],4)) for dic in self.data.get('ml_emotion', {}).get('sentence', {}) if dic['value'] != 0]
        elif self.__class__.__name__ == 'Subsentence':
            return [(dic['type'], round(dic['value'],4)) for dic in self.data.get('ml_emotion', {}).get('subsentence', {}) if dic['value'] != 0]
            

    @DictProperty
    def sentiment(self) -> dict:
        return self.data.get('sentiment', {}).get('values', {})

    @DictProperty
    def sentiment_ml(self) -> dict:
        if self.__class__.__name__ == 'Sentence':
            return {'total': self.data.get('ml_sentiment', {}).get('sentence', {}).get('value', 0)}
        elif self.__class__.__name__ == 'Subsentence':
            return {'total': self.data.get('ml_sentiment', {}).get('subsentence', {}).get('value', 0)}

    @StrProperty
    def sentence_type(self) -> str:
        return self.data.get('sentence_type', None)

    @ListProperty
    def token(self) -> list:
        return [s.get('source', None) for s in self.data.get('detail', [])]
    
    @ListProperty
    def spans(self) -> list:
        spans = [self.ref_document.document_data.get('coreference').get('spans', [])[idx] for idx in self.data.get('coreference', [])]
        return [Span(sp, self.ref_document) for sp in spans]

    @ListProperty
    def clusters(self) -> list:
        spans = [self.ref_document.document_data.get('coreference').get('spans', [])[idx] for idx in self.data.get('coreference', [])]
        clusters_idx = list(set([span['cluster_index'] for span in spans]))
        clusters = [(cl, self.ref_document.document_data.get('coreference').get('clusters', [])[cl]) for cl in clusters_idx]
        return [Cluster(cluster_idx, spans_idx, self.ref_document) for cluster_idx, spans_idx in clusters]


from .utils import flatten_lst, StrProperty, ListProperty, DictProperty, IntProperty
from .TextChunk import TextChunk
from .utils import GLOBAL, DOC, SENT, SUB, TOK
from .Sentence import Sentence
from .Cluster import Cluster
from .Span import Span

class Document(TextChunk):
    __slots__ = ("sentences", "document_data", "n", "max", "_id")
    
    def __init__(self, document_data, _id):
        """Class for the document, a collection of sentences with
            additional information at the document level.
        Args:
            document_data (dict): dict with document data.
            _id (int or str): id of the document.
        """
        super(Document, self).__init__()
        self.sentences = [Sentence(s, i, self) for i, s in enumerate(document_data['sentences'])]
        self.document_data = {k:v for k,v in document_data.items() if k != 'sentences'}
        self.max = len(self.sentences)
        self.data = self.sentences
        self._id = str(_id)

    @property
    def id(self):
        return self._id
    
    def __repr__(self):
        return str(self.sentences)

    def _get_data(self) -> dict:
        tmp = self.document_data
        tmp['sentences'] = [s.data for s in self.sentences]
        return tmp

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> Sentence:
        if self.n < self.max:
            self.n += 1
            return self.sentences[self.n - 1]
        else:
            raise StopIteration
    
    @ListProperty
    def subsentences(self) -> list:
        return flatten_lst([s.subsentences for s in self.sentences])

    @ListProperty
    def tokens(self) -> list:
        return flatten_lst([s.tokens for s in self.sentences])

    @ListProperty
    def documents(self) -> list:
        return [self]

    @StrProperty
    def str_doc(self) -> str:
        return self.document_data.get('source_pure', None)

    @StrProperty
    def original_text_doc(self) -> str:
        return self.document_data.get('source_pure', None)

    @DictProperty
    def emotion_doc(self) -> list:
        return self.document_data.get('emotion', None)
    
    @IntProperty
    def sentiment_doc(self) -> float:
        return self.document_data.get('sentiment', None)
    
    @StrProperty
    def domain_doc(self):
        return self.document_data.get('domain', None)

    @StrProperty
    def type_doc(self):
        return self.document_data.get('type', None)

    @StrProperty
    def original_text_doc(self) -> str:
        return self.document_data.get('source_pure', None)

    @ListProperty
    def emoticon_doc(self) -> list:
        return self.document_data.get('emoticon_data', {})

    @ListProperty
    def spans(self) -> list:
        return [Span(sp, self) for i, sp in enumerate(self.document_data.get('coreference').get('spans', []))]

    @ListProperty
    def clusters(self) -> list:
        return [Cluster(i, spans_idx, self) for i, spans_idx in enumerate(self.document_data.get('coreference').get('clusters', []))]

    def replace_coreference(self, attribute = 'source', replace=['CLS']) -> list:
        """Replaces coreference mentions with the head of the cluster in the text

        Args:
            attribute (str, optional): Attribute to get. Defaults to 'source'.
            replace (list, optional): Defines what kind of spans get replaced by head span. Defaults to ['CLS'].

        Returns:
            list: List of sentences with the desired attribute information and replacement.
        """
        text = [[getattr(t, attribute) for t in s ] for s in self.sentences]
        for cl in self.clusters:
            head = cl.head
            children = sorted(cl.children, key=lambda x: x.tokens_idx[-1], reverse=True)
            for child in children:
                if set(replace) & set(child.get_attributes('pos')):
                    text[child.sentence_idx] = text[child.sentence_idx][:child.tokens_idx[0]] \
                    + text[head.sentence_idx][head.tokens_idx[0]:head.tokens_idx[-1] + 1] \
                    + text[child.sentence_idx][child.tokens_idx[-1] + 1:]
        return text


import sys
import os
import json

import jsonlines as jsonl

from .client import Client
from .utils import flatten_lst, StrProperty, ListProperty, DictProperty, IntProperty
from .TextChunk import TextChunk
from .utils import GLOBAL, DOC, SENT, SUB, TOK
from .Document import Document
from .Sentence import Sentence
from .Subsentence import Subsentence
from .Token import Token

class PipelineError(Exception): pass

class NLP(TextChunk):

    def __init__(self, api_key = None, client = None, verbose=True):
        """Class for data analysis of API return.
        Provides both high and low level methods to access data via specific class
        for each key in the api result or methods designed for specific use cases.

        Args:
            api_key (str, optional): api_key used for authentification to lettria's API. Defaults to None.
            client (Client, optional): Client class used to make api requests.. Defaults to None.
            verbose (bool, optional): Turns on/off verbosity. Defaults to True.
        """
        super(NLP, self).__init__()
        self.client = None
        if client or api_key:
            self.add_client(client, api_key)
        self.documents = []
        self.max = len(self.documents)
        self.data = self.documents
        self._next_id = 0

        self.token_fields = [p for p in dir(Token) if isinstance(getattr(Token,p),property)]
        self.sentence_fields = []
        self.document_fields = []
        if 'token_flat' not in dir(NLP):
            self._generate_properties()
        else:
            self.sentence_fields = [p for p in dir(Sentence) if isinstance(getattr(Sentence,p),property) and not p.endswith('flat')]
            self.document_fields = [p for p in dir(Document) if isinstance(getattr(Document,p),property)  and not p.endswith('flat')]
        self.fields = list(set(self.sentence_fields + self.document_fields))

        def doNothing(*args):
            pass
        global print
        if not verbose:
            print = doNothing

    @ListProperty
    def sentences(self):
        """ Direct access to all sentences """
        return flatten_lst([d.sentences for d in self.documents])

    @ListProperty
    def subsentences(self):
        """ Direct access to all subsentences """
        sentences = flatten_lst([d.sentences for d in self.documents])
        return flatten_lst([s.subsentences for s in sentences])

    @ListProperty
    def tokens(self):
        """ Direct access to all tokens """
        return flatten_lst([s.tokens for s in self.sentences])

    @property
    def nlp(self):
        return [self]

    def _generate_properties(self):
        """ Dynamically create properties for NLP, Document, Sentence and Subsentence classes to avoid
            code redudancies.
            Methods are implemented at the level at which the information is available,
            wrapper functions at higher levels are added dynamically."""

        #add wrapper methods for tokens to sentence and subsentence classes
        for field in self.token_fields:
            for _class in [Sentence, Subsentence]:
                if field in ['str']:
                    continue
                if not hasattr(_class, field):
                    setattr(_class, field, ListProperty(self._make_lambda(field, token=True)))
                if not hasattr(_class, field + '_flat'):
                    setattr(_class, field + '_flat', ListProperty(self._make_lambda(field, True, token=True)))

        #update list of sentence properties
        self.sentence_fields = [p for p in dir(Sentence) if isinstance(getattr(Sentence,p),property) and not p.endswith('flat')]

        #add wrapper methods to Document and Subsentence
        for field in self.sentence_fields:
            for _class in [Document]:
                    if field in ['subsentences', 'tokens', 'sentences']:
                        continue
                    if not hasattr(_class, field):
                        setattr(_class, field, ListProperty(self._make_lambda(field)))
                    if not hasattr(_class, field + '_flat'):
                        setattr(_class, field + '_flat', ListProperty(self._make_lambda(field, True)))
        
            #add methods to Subsentence class
            filters = ['token', 'str', 'detail', \
                            'sentiment', 'sentiment_ml', 'emotion', 'emotion_ml', 'tokens']
            for _class in [Subsentence]:
                if field in filters:
                    if not hasattr(_class, field):
                        setattr(_class, field, getattr(Sentence, field))
                elif field not in ['subsentences']:
                    if not hasattr(_class, field):
                        setattr(_class, field, property(self._make_lambda_not_available(field)))

            #adding _flat  variants to sentence and subsentence methods for convenience
            for _class in [Sentence, Subsentence]: 
                if not hasattr(_class, field + '_flat'):
                    setattr(_class, field + '_flat', ListProperty(self._make_lambda_flat(field)))

        #update list of document properties
        self.document_fields = [p for p in dir(Document) if isinstance(getattr(Document,p),property)  and not p.endswith('flat')]

        #add wrapper methods to NLP
        for field in self.document_fields:
            for _class in [NLP]:
                if field in ['documents']:
                    continue
                if not hasattr(_class, field):
                    setattr(_class, field, property(self._make_lambda(field)))
                if not hasattr(_class, field + '_flat'):
                    setattr(_class, field + '_flat', property(self._make_lambda(field, True)))

    def _make_lambda(self, field, flatten = False, token=False):
        """ Returns lambda functions to be added as property to classes """
        if token:
            if not flatten:
                return lambda c_self : [getattr(t, field) for t in c_self.tokens]
            else:
                return lambda c_self : flatten_lst([getattr(t, field) for t in c_self.tokens])
        else:
            if not flatten:
                return lambda c_self : [getattr(s, field) for s in c_self.data]
            else:
                return lambda c_self : flatten_lst([getattr(s, field) for s in c_self.data])



    def _make_lambda_not_available(self, field):
        """ Returns lambda functions to be added as property to classes """
        return lambda x: field + ' not available for subsentences'

    def _make_lambda_flat(self, field):
        """ Returns lambda functions to be added as property to classes """
        return lambda c_self: flatten_lst(getattr(c_self, field))

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < self.max:
            self.n += 1
            return self.documents[self.n - 1]
        else:
            raise StopIteration

    def add_client(self, client = None, api_key = None):
        """ Add/Change the active Client by providing a Client Instance or an api_key
            Args:
                client: Client Instance
                api_key: string representing a valid api_key
        """
        if client and isinstance(client, Client):
            self.client = client
        elif api_key:
            self.client = Client(api_key)
        else:
            print('Please provide a client instance or an api_key.')

    def _missing_api_key():
        print('Failure : Please assign a client or api_key with add_client() in order to make a request. ')

    def _preprocess_document(self, document):
        if isinstance(document, str):
            if document.strip():
                document = document.strip()
                # document = [document] ## Temporary until option split_sentence is available
            else:
                return " "
        elif isinstance(document, list):
            document = [d.strip() for d in flatten_lst(document) if d.strip()]
            document = ' '.join(document) ## Temporary until option split_sentence is available
        else:
            print('document argument should be of type list or str.')
            raise TypeError
        return document

    def add_documents(self, documents, batch_size=32, skip_document=False, document_ids=[], verbose=True):
        """ Performs request to lettria API for multiples documents and stores it.
            This will increase speed by allowing batching to occur across different documents.
            Args:
                skip_document: If False document is added even if empty or request failed.
                batch_size: Number of documents to request at once.
                document_ids: List of Ids given to documents, by default sequential integer """
        if not isinstance(documents, list):
            print("ERROR, 'add_documents' method expects a list of string as first argument.")
            return
        if document_ids:
            if not isinstance(document_ids, list) or len(document_ids) == 0 or len(document_ids) != len(documents):
                print("ERROR, 'document_ids' argument should be a list of string with the same length as the number of documents")
                return
        
        if not self.client:
            self._missing_api_key()
            return
        
        for idx_batch in range(0, len(documents), batch_size):
            batch_documents = documents[idx_batch:idx_batch + batch_size]
            batch_documents = [self._preprocess_document(doc) for doc in batch_documents]
            results = []
            if not batch_documents:
                print("ERROR, documents input is empty.")
                return 
            results, document_ids = self.client.request_batch_documents(batch_documents, document_ids = document_ids, skip_document=skip_document)
            for idx, (input_document, result) in enumerate(zip(batch_documents, results)):
                if idx_batch + idx < len(document_ids):
                    next_doc_id = document_ids[idx_batch + idx]
                else:
                    next_doc_id = self._next_id
                    self._next_id += 1
                if isinstance(result, dict) and result:
                    self.documents.append(Document(result, _id=next_doc_id))
                    if verbose:
                        print("Added document " + str(self.documents[-1].id) + '.')
                else:
                    if not input_document:
                        self.documents.append(Document([], _id=next_doc_id))
                        if verbose:
                            print("Added empty document " + str(self.documents[-1].id) + ': received empty input.')
                    else:
                        self.documents.append(Document([], _id=next_doc_id))
                        if verbose:
                            print("Added empty document " + str(self.documents[-1].id) + ': processing failed.')
                self.max += 1

    def add_document(self, document, skip_document=False, _id=None, verbose=True):
        """ Performs request to lettria API for a document and stores it.
            Args:
                skip_document: If False document is added even if empty or request failed.
                id: Id given to document, by default sequential integer """
        if not self.client:
            self._missing_api_key()
            return
        results = None
        document = self._preprocess_document(document)
        
        if document:
            results = self.client.request_document(document, skip_document=skip_document, verbose=verbose)
        if results is None and skip_document == True:
            if not document:
                print("Skpping document, received empty input.")
            else:
                print("Skpping document, processing failed.")
        else:
            if _id == None:
                next_doc_id = self._next_id
                self._next_id += 1
            else:
                next_doc_id = _id
            if isinstance(results, dict):
                self.documents.append(Document(results, _id=next_doc_id))
                if verbose:
                    print("Added document " + str(self.documents[-1].id) + '.')
            else:
                if not document:
                    self.documents.append(Document([], _id=next_doc_id))
                    print("Added empty document " + str(self.documents[-1].id) + ': received empty input.')
                else:
                    self.documents.append(Document([], _id=next_doc_id))
                    print("Added empty document " + str(self.documents[-1].id) + ': processing failed.')
            self.max += 1

    def add_document_data(self, doc_data, _id=None):
        self.documents.append(Document(doc_data, _id))
        self.max += 1

    def _get_data(self):
        return [d._get_data() for d in self.documents]

    def save_result(self, *args):
        """ Alias for save_result"""
        self.save_results(*args)

    def save_results(self, filename = ''):
        """ Writes json result to a file with the specified name."""
        path_ok = 0
        c = 0
        if not self.data:
            print("No data to save.")
            return
        if not filename:
            filename = 'results'
            while not path_ok:
                path = filename + '_' + str(c) + '.jsonl'
                if not os.path.isfile(path):
                    path_ok = 1
                else:
                    c += 1
        else:
            path = filename
            if path.endswith('.jsonl'):
                path = path[:-6]
            path = path + '.jsonl'
        try:
            with jsonl.open(path, 'w') as fw:
                for d in self.documents:
                    fw.write({'document_id':d.id, 'data':d._get_data()})
            print(f'Results saved to {path}')
        except Exception as e:
            print(e)

    def load_result(self, *args, **kwargs):
        """ Alias for load_results"""
        self.load_results(*args, **kwargs)

    def load_results(self, path = 'results_0', reset = False, chunksize = None, verbose=True):
        """ Loads result from a valid json file."""
        if path.endswith('.json') or path.endswith('.jsonl'):
            pass
        else:
            path = path + '.jsonl'
        try:
            if reset:
                self.reset_data()
            if path.endswith('jsonl'):
                with jsonl.open(path, 'r') as f:
                    for line in f:
                        self.add_document_data(line.get('data'), _id=line.get('document_id', None))
            else:
                raise Exception("Expected jsonl file extension.")
            if verbose:
                print(f'Loaded {str(path)} successfully')
        except Exception as e:
            print('Failure to load ' + str(path) + ': ', e)

    def split_results(self, input_file, max_document_per_file, output_file=''):
        """ Splits existing json results into multiple files """
        if not output_file:
            output_file = input_file
        with open(input_file, 'r') as f:
            results = json.load(f)
        i = 0
        count = 0
        while i < len(self.documents):
            if isinstance(results, list):
                with open(f"{count}_{output_file}", 'w') as fw:
                    json.dump(result[i:i + max_document_per_file], fw)
            elif isinstance(results, dict):
                with open(f"{count}_{output_file}", 'w') as fw:
                    json.dump({'document_ids':results['document_ids'][i:i + max_document_per_file],\
                            'documents':results['documents'][i:i + max_document_per_file]}, fw)
            i += max_document_per_file
            count += 1
            results['document_ids']

    def reset_data(self):
        """ Erase current data """
        self.documents = []
        self.data = self.documents
        self.max = 0
        self._next_id = 0
    
    def labels_to_annotation(self, labels_list, description_list = None, parent_list = None, output_file=''):
        if not description_list:
            description_list = ['' for k in labels_list]
        if not parent_list:
            parent_list = [None for k in labels_list]
        labels_json = [{'label': label, 'description':desc, 'parent':parent} for label, desc, parent in zip(labels_list, description_list, parent_list)]
        if output_file:
            with open(output_file, 'w') as fw:
                json.dump(labels_json, fw)
        return labels_json
        
    def to_annotation_format(self, output_file, attribute=None, filter_list = [], verbose=True):
        """ Writes data to a file in the annotation format for lettria's platform.

        Args:
            output_file (str): Name of the file to write to.
            attribute (str, optional): Attribute to be used to preselect tokens to annotate. Defaults to None.
            filter_list (list, optional): Filter used to compare to the chosen attribute. Defaults to [].
            verbose (bool, optional): Turns off/on verbosity. Defaults to True.
        """
        if filter_list and attribute is None:
            print("filter_list argument is not empty but no attribute has been chosen.")
            return
        if not output_file.endswith('.json'):
            output_file = output_file + '.json'
        if not self.data:
            print("No data to save.")
            return
        try:
            data = []
            for doc in self.documents:
                tmp = {'items':[], 'annotate':[]}
                token_count = 0
                for s in doc:
                    tmp['items'] += [t.str for t in s.tokens] + ['\n']
                    if attribute is not None:
                        tmp['annotate'] += [i + token_count for i, t in enumerate(s.tokens) if getattr(t, attribute) in filter_list]
                    token_count += len([t.str for t in s.tokens] + ['\n'])
                data.append(tmp)
            with open(output_file, 'w') as fw:
                json.dump(data, fw)
            if verbose:
                print(f'Annotations saved to {output_file}')
        except Exception as e:
            print(e)
    
    def reformat_data(self, filepath, output_file):
        """ Takes the path of a results file with the old sentence format 
            and writes it to a new file with the new document format
            
            Args:
                filepath (str): Name of the file to read data from.
                output_file (str): Name of the file to write to.
            """
        data = []
        idx = 0
        if filepath.endswith('jsonl'):
            with jsonl.open(filepath, 'r') as f:
                for l in f:
                    data.append(l)
        elif filepath.endswith('json'):
            with open(filepath, 'r') as f:
                tmp = json.load(f)
                for t in tmp:
                    data.append({'id':idx, 'data':t})
                idx += 1
        new_data = []
        new_format = {
            'coreference': None,
            'type': None, 
            'domain': None,
            'source_pure': None,
            'language': None,
            'emoticon': None,
            'sentiment': None,
            'emotion': None,
            'sentences': None
        }
        ids = []
        for batch in data:
            tmp = new_format.copy()
            tmp_sentences = []
            ids.append(batch.get('document_id', None))
            tmp['source_pure'] = ' '.join([d.get('source_pure', d.get('source', '')) for d in batch['data']])
            tmp['language'] = batch['data'][0].get('language_used', {}).get('predict', 'fr')
            for s in batch['data']:
                new_sentence = {}
                new_sentence['source'] = s.get('source', None)
                new_sentence['source_pure'] = s.get('source_pure', None)
                new_sentence['sentence_indexes'] = s.get('sentence_source_indexes', None)
                new_sentence['subsentences'] = s.get('proposition', None)
                new_sentence['emotion'] = s.get('emotion', None)
                new_sentence['sentiment'] = s.get('sentiment', None)
                new_sentence['sentence_type'] = s.get('sentence_acts', {}).get('predict', None)
                new_sentence['ner'] = s.get('NER', None)
                new_sentence['detail'] = s.get('synthesis', None)
                tmp_sentences.append(new_sentence)
            tmp['sentences'] = tmp_sentences
            new_data.append(tmp)
        if new_data:
            with jsonl.open(output_file, 'w') as f:
                for _id, d in zip(ids, new_data):
                    f.write({'data':d, 'document_id':_id})


from .utils import StrProperty, ListProperty, DictProperty
from .transform_morph import transform_data

class Token:
    __slots__ = ("data", "idx", "pure_source", "ref_sentence")

    def __init__(self, data, idx, pure_source, ref_sentence):
        """Class for a single token, usually a word or symbol.
        Args:
            data (dict): Dict with the data of the token.
            idx (int): Index of the token in the sentence.
            pure_source (str): Original text of the token.
            ref_sentence (Sentence): Reference to the Sentence object.
        """
        self.data = data
        self.idx = idx
        self.pure_source = pure_source
        self.ref_sentence = ref_sentence

    def __repr__(self):
        return self.str

    def __str__(self) -> str:
        return self.data.get('source', None)

    @StrProperty
    def token(self) -> str:
        return self.data.get('source', None)

    @StrProperty
    def original_text(self) -> str:
        idx = self.data.get('source_indexes', None)
        if idx and len(idx) == 2:
            return self.pure_source[idx[0]:idx[1]]
        else:
            return self.token

    @StrProperty
    def source(self) -> str:
        return self.data.get('source', None)

    @StrProperty
    def str(self) -> str:
        return self.data.get('source', None)

    @StrProperty
    def pos(self) -> str:
        return self.data.get('tag', None)

    @DictProperty
    def ner(self) -> dict:
        type_ = self.data.get('type', '')
        value = self.data.get('value', '')
        return {'type': type_, 'value': value} if type_ else {}

    @ListProperty
    def spans(self) -> list:
        if self.data.get('coreference', []):
            return [span for span in self.ref_sentence.spans if self.idx in span.tokens_idx]
        return []

    @ListProperty
    def clusters(self) -> list:
        if self.data.get('coreference', []):
            return [cluster for cluster in self.ref_sentence.clusters if set(cluster.spans_idx) & set(self.data.get('coreference'))]
        return []

    @StrProperty
    def lemma(self) -> str:
        return self.data.get('lemma', None)

    @ListProperty
    def lemma_detail(self) -> list:
        if 'lemmatizer' in self.data:
            if isinstance(self.data['lemmatizer'], list):
                return self.data['lemmatizer']
            elif isinstance(self.data['lemmatizer'], dict):
                return [self.data['lemmatizer']]
        else:
            return []

    @DictProperty
    def auxiliary(self) -> dict:
        return self.data.get('auxiliary', {})

    @StrProperty
    def gender(self) -> str:
        l_d = self.lemma_detail
        if len(l_d) != 1:
            return None
        else:
            gdr = l_d[0].get('gender', {}).get('female', None)
            return 'feminine' if gdr == True  else 'masculine' if gdr != None else None

    @StrProperty
    def plural(self) -> str:
        l_d = self.lemma_detail
        if len(l_d) != 1:
            return None
        else:
            plr = l_d[0].get('gender', {}).get('plural', None)
            return 'plural' if plr == True  else 'singular' if plr != None else None

    @ListProperty
    def infinitive(self) -> list:
        lst_inf = [detail.get('infinit', None) for detail in self.lemma_detail if detail.get('infinit', None)]
        return lst_inf 

    @StrProperty
    def mode(self) -> str:
        lst_mode = []
        for detail in self.lemma_detail:
            if 'mode' in detail:
                return detail['mode']
        return None

    @ListProperty
    def conjugate(self) -> list:
        lst = []
        for detail in self.lemma_detail:
            if 'conjugate' in detail or 'infinit' in detail:
                lst += [{'infinit': detail.get('infinit', None), **d} for d in detail.get('conjugate', [])]
        return lst

    @StrProperty
    def morphology(self) -> str:
        if self.data.get('lemmatizer', []):
            return transform_data(self.data['lemmatizer'], self.data['source'])
        else:
            return ''

    @StrProperty
    def dep(self) -> str:
        return self.data.get('dep', None)

    @StrProperty
    def ref(self) -> int:
        return self.data.get('ref', None)

    @ListProperty
    def meaning(self) -> list:
        return [(m.get('super', ''), m.get('sub', '')) for m in self.data.get('meaning', [])]

    def __format__(self, format_spec):
        return format(str(self), format_spec)



from .Span import Span

hierarchy_pos_cluster = ['NP', ['PD','N'], ['D', 'N'], 'N', 'CLS', 'CLO', 'PRON', 'D', 'PROREL', 'PD', 'P', 'ENTITY'] 

class Cluster():
    __slots__ = ["cluster_idx", "spans_idx", "ref_document", "spans", "n"]

    def __init__(self, cluster_idx, spans_idx, ref_document):
        """
        Class for coreference cluster.
        Args:
            cluster_idx (int): Index of the cluster in the document.
            spans_idx (List): Indexes of the spans of the cluster.
            ref_document (Document): Reference to the Document object
        """
        self.spans_idx = spans_idx
        self.cluster_idx = cluster_idx
        self.ref_document = ref_document
        self.spans = [Span(self.ref_document.document_data.get('coreference', {}).get('spans', {})[span_idx], ref_document) for span_idx in self.spans_idx]
        
    def __str__(self):
        return ', '.join([str(span) for span in self.spans])

    def __repr__(self):
        return str(self)
    
    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.spans):
            self.n += 1
            return self.spans[self.n - 1]
        else:
            raise StopIteration
    
    @property
    def head(self) -> Span:
        """Returns the head of a cluster, which is the span that best represent the cluster.
        This is done according to a hierarchy that use the POS tags inside the spans.

        Returns:
            Span: Span of the head of the cluster
        """
        for h in hierarchy_pos_cluster:
            if isinstance(h, list):
                match = [s for s in self.spans if set(h) & set(s.get_attributes('pos')) == set(h)]
            else:
                match = [s for s in self.spans if h in s.get_attributes('pos')]
            if match:
                return match[0]
        return self.spans[0]

    @property
    def children(self) -> list:
        """Returns the children of the cluster, i.e. all spans except head.
        Returns:
            list: list of children spans.
        """
        head = self.head
        return [span for span in self.spans if span != head]


import os

debug = os.getenv('DEBUG') == '1' or os.getenv('DEBUG') == 'True'
flatten = lambda l: sum(map(flatten,l),[]) if isinstance(l,list) else [l]

#Accepted arguments for hierarchical levels
GLOBAL =    ['g', 'global', 'glob']
DOC =       ['d', 'doc', 'document', 'documents']
SENT =      ['s', 'sentence', 'sent', 'sentences']
SUB =       ['sub', 'subsentence', 'subsentences']
TOK =       ['t', 'token', 'tok', 'tokens']

POSITIVE = ['positive', 'positif', 'pos', '+']
NEGATIVE = ['negative', 'negatif', 'neg', '-']
NEUTRAL = ['neutral', 'neutre', 'neut']
EMOTIONS = ['admiration', 'amusement', 'anger', 'annoyance', 'caring', 'confusion', 'curiosity', 'desire',
    'disappointment', 'disapproval', 'disgust', 'embarrassment', 'excitement', 'fear', 'gratitude',
    'grief', 'happiness', 'love', 'nervousness', 'optimism', 'pride', 'realization', 'relief', 'remorse', 'sadness', 'surprise']
SENTENCE_TYPES = ['command', 'assert', 'question_open', 'question_closed']


def flatten_lst(lst):
    return flatten(lst)

class StrProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except Exception as e:
            if debug:
                raise e
            else:
                return None

class IntProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except Exception as e:
            if debug:
                raise e
            else:
                return 0

class ListProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except Exception as e:
            if debug:
                raise e
            else:
                return []

class DictProperty(property):
    def __get__(self, obj, objtype=None):
        try:
            if obj is None:
                return self
            if self.fget is None:
                raise AttributeError("unreadable attribute")
            return self.fget(obj)
        except Exception as e:
            if debug:
                raise e
            else:
                return {}


import requests
import random
import json

class RequestError(Exception): pass

class Client:
    def __init__(self, key=None):
        if not key:
            raise Exception('Missing key for initialization')
        if isinstance(key, str) is False:
            raise Exception('Wrong key format')
        if key.startswith('LettriaProKey'):
            key = key[14:]
        self.key = key
        self.headers = { 'Authorization': 'LettriaProKey ' + str(self.key), 'content-type': 'application/json' }
        self.max_try = 2

    def print_response_error(self, response):
        if 'Error' in response:
            print("Lettria SDK: \033[1;31;40mERROR:\033[0;37;40m {}".format(response['Error']))
        else:
            print(response)

    def request(self, text):
        result = None
        response = None
        i = 0
        while i < self.max_try:
            try:
                response = requests.post('https://api.lettria.com/', headers=self.headers, json={'text' : text}).json()
                if not response or (response and not isinstance(response, dict)):
                    raise Exception
                result = response
                break
            except Exception as e:
                print(e)
                i += 1
        if result is None:
            print(f'Request failed after {self.max_try} tries.')
        if result and not isinstance(result, dict):
            result = None
        return result
    
    def request_batch(self, batch_documents):
        result = []
        i = 0
        while i < self.max_try:
            try:
                response = requests.post('https://api.lettria.com/', headers=self.headers, json={'documents' : batch_documents}).json()
                if not response or (response and not isinstance(response, list)):
                    raise Exception
                result = response
                break
            except Exception as e:
                i += 1
        if result is None or len(result) == 0:
            print(f'Batch request failed after {self.max_try} tries.')
        if result and not isinstance(result, list):
            result = []
        return result
    
    def _check_result(self, result):
        """ Check if detail key is empty """
        bad_idx = []
        for i, r in enumerate(result):
            if not isinstance(r, dict) or not 'detail' in r:
                bad_idx.append(i)
        return bad_idx
    
    def request_document(self, document, skip_document = False, verbose=False):
        """ Input: string or list of string"""
        results = []
        try:
            results = self.request(document)
            if results is None:
                raise RequestError
        except RequestError:
            if skip_document:
                return None
        return results

    def request_batch_documents(self, batch_documents, document_ids, skip_document = False):
        ''' Input: string or list of string'''
        results = self.request_batch(batch_documents)
        if skip_document:
            for idx in range(len(results) - 1, -1, -1):
                if not results[idx]:
                    print("WARNING: Skipping document, request error.")
                    results.pop(idx)
                    if document_ids:
                        document_ids.pop(idx)
        return results, document_ids


from .transform_morph import transform_data
from .utils import StrProperty, ListProperty, DictProperty
from .TextChunk import TextChunk
from .Token import Token

class Subsentence(TextChunk):
    __slots__ = ("data", "n", "max")

    def __init__(self, data_sentence):
        """Class for a subsentence, a group of words part of a sentence.
        Args:
            data_sentence (dict): Dict with the data relative to the subsentence.
        """
        super(Subsentence, self).__init__()
        self.data = data_sentence
        self.max = len(self.data['detail'])

    def __repr__(self):
        return self.str

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> Token:
        if self.n < self.max:
            self.n += 1
            return Token(
                self.data.get('detail', [])[self.n - 1], 
                self.n -1, 
                self.data.get('source_pure', self.data.get('source', None)), 
                self)
        else:
            raise StopIteration
    
    @ListProperty
    def subsentences(self) -> list:
        return [self]